@@ -20,6 +20,7 @@
 
 #define PY_SSIZE_T_CLEAN
 #include "Python.h"
+#include "thirdparty/pythoncapi_compat.h"
 #include "libImaging/Imaging.h"
 
 #include <ft2build.h>
@@ -47,17 +48,17 @@
     ;
 
 #ifdef HAVE_RAQM
-# ifdef HAVE_RAQM_SYSTEM
-#  include <raqm.h>
-# else
-#  include "thirdparty/raqm/raqm.h"
-#  ifdef HAVE_FRIBIDI_SYSTEM
-#   include <fribidi.h>
-#  else
-#   include "thirdparty/fribidi-shim/fribidi.h"
-#   include <hb.h>
-#  endif
-# endif
+#ifdef HAVE_RAQM_SYSTEM
+#include <raqm.h>
+#else
+#include "thirdparty/raqm/raqm.h"
+#ifdef HAVE_FRIBIDI_SYSTEM
+#include <fribidi.h>
+#else
+#include "thirdparty/fribidi-shim/fribidi.h"
+#include <hb.h>
+#endif
+#endif
 #endif
 
 static int have_raqm = 0;
@@ -81,6 +82,9 @@ struct {
     /* font objects */
 
     static FT_Library library;
+#ifdef Py_GIL_DISABLED
+static PyMutex ft_library_mutex;
+#endif
 
 typedef struct {
     PyObject_HEAD FT_Face face;
@@ -125,7 +129,8 @@ getfont(PyObject *self_, PyObject *args, PyObject *kw) {
     unsigned char *font_bytes;
     Py_ssize_t font_bytes_size = 0;
     static char *kwlist[] = {
-        "filename", "size", "index", "encoding", "font_bytes", "layout_engine", NULL};
+        "filename", "size", "index", "encoding", "font_bytes", "layout_engine", NULL
+    };
 
     if (!library) {
         PyErr_SetString(PyExc_OSError, "failed to initialize FreeType library");
@@ -147,7 +152,8 @@ getfont(PyObject *self_, PyObject *args, PyObject *kw) {
             &encoding,
             &font_bytes,
             &font_bytes_size,
-            &layout_engine)) {
+            &layout_engine
+        )) {
         PyConfig_Clear(&config);
         return NULL;
     }
@@ -165,7 +171,8 @@ getfont(PyObject *self_, PyObject *args, PyObject *kw) {
             &encoding,
             &font_bytes,
             &font_bytes_size,
-            &layout_engine)) {
+            &layout_engine
+        )) {
         return NULL;
     }
 #endif
@@ -183,7 +190,9 @@ getfont(PyObject *self_, PyObject *args, PyObject *kw) {
 
     if (filename && font_bytes_size <= 0) {
         self->font_bytes = NULL;
+        MUTEX_LOCK(&ft_library_mutex);
         error = FT_New_Face(library, filename, index, &self->face);
+        MUTEX_UNLOCK(&ft_library_mutex);
     } else {
         /* need to have allocated storage for font_bytes for the life of the object.*/
         /* Don't free this before FT_Done_Face */
@@ -193,12 +202,15 @@ getfont(PyObject *self_, PyObject *args, PyObject *kw) {
         }
         if (!error) {
             memcpy(self->font_bytes, font_bytes, (size_t)font_bytes_size);
+            MUTEX_LOCK(&ft_library_mutex);
             error = FT_New_Memory_Face(
                 library,
                 (FT_Byte *)self->font_bytes,
                 font_bytes_size,
                 index,
-                &self->face);
+                &self->face
+            );
+            MUTEX_UNLOCK(&ft_library_mutex);
         }
     }
 
@@ -233,18 +245,6 @@ getfont(PyObject *self_, PyObject *args, PyObject *kw) {
     return (PyObject *)self;
 }
 
-static int
-font_getchar(PyObject *string, int index, FT_ULong *char_out) {
-    if (PyUnicode_Check(string)) {
-        if (index >= PyUnicode_GET_LENGTH(string)) {
-            return 0;
-        }
-        *char_out = PyUnicode_READ_CHAR(string, index);
-        return 1;
-    }
-    return 0;
-}
-
 #ifdef HAVE_RAQM
 
 static size_t
@@ -254,7 +254,8 @@ text_layout_raqm(
     const char *dir,
     PyObject *features,
     const char *lang,
-    GlyphInfo **glyph_info) {
+    GlyphInfo **glyph_info
+) {
     size_t i = 0, count = 0, start = 0;
     raqm_t *rq;
     raqm_glyph_t *glyphs = NULL;
@@ -266,28 +267,37 @@ text_layout_raqm(
         goto failed;
     }
 
+    Py_ssize_t size;
+    int set_text;
     if (PyUnicode_Check(string)) {
         Py_UCS4 *text = PyUnicode_AsUCS4Copy(string);
-        Py_ssize_t size = PyUnicode_GET_LENGTH(string);
+        size = PyUnicode_GET_LENGTH(string);
         if (!text || !size) {
             /* return 0 and clean up, no glyphs==no size,
                and raqm fails with empty strings */
             goto failed;
         }
-        int set_text = raqm_set_text(rq, text, size);
+        set_text = raqm_set_text(rq, text, size);
         PyMem_Free(text);
-        if (!set_text) {
-            PyErr_SetString(PyExc_ValueError, "raqm_set_text() failed");
+    } else if (PyBytes_Check(string)) {
+        char *buffer;
+        PyBytes_AsStringAndSize(string, &buffer, &size);
+        if (!buffer || !size) {
+            /* return 0 and clean up, no glyphs==no size,
+               and raqm fails with empty strings */
             goto failed;
         }
-        if (lang) {
-            if (!raqm_set_language(rq, lang, start, size)) {
-                PyErr_SetString(PyExc_ValueError, "raqm_set_language() failed");
-                goto failed;
-            }
-        }
+        set_text = raqm_set_text_utf8(rq, buffer, size);
     } else {
-        PyErr_SetString(PyExc_TypeError, "expected string");
+        PyErr_SetString(PyExc_TypeError, "expected string or bytes");
+        goto failed;
+    }
+    if (!set_text) {
+        PyErr_SetString(PyExc_ValueError, "raqm_set_text() failed");
+        goto failed;
+    }
+    if (lang && !raqm_set_language(rq, lang, start, size)) {
+        PyErr_SetString(PyExc_ValueError, "raqm_set_language() failed");
         goto failed;
     }
 
@@ -302,13 +312,14 @@ text_layout_raqm(
 #if !defined(RAQM_VERSION_ATLEAST)
             /* RAQM_VERSION_ATLEAST was added in Raqm 0.7.0 */
             PyErr_SetString(
-                PyExc_ValueError,
-                "libraqm 0.7 or greater required for 'ttb' direction");
+                PyExc_ValueError, "libraqm 0.7 or greater required for 'ttb' direction"
+            );
             goto failed;
 #endif
         } else {
             PyErr_SetString(
-                PyExc_ValueError, "direction must be either 'rtl', 'ltr' or 'ttb'");
+                PyExc_ValueError, "direction must be either 'rtl', 'ltr' or 'ttb'"
+            );
             goto failed;
         }
     }
@@ -328,29 +339,23 @@ text_layout_raqm(
         len = PySequence_Fast_GET_SIZE(seq);
         for (j = 0; j < len; j++) {
             PyObject *item = PySequence_Fast_GET_ITEM(seq, j);
-            char *feature = NULL;
-            Py_ssize_t size = 0;
-            PyObject *bytes;
-
             if (!PyUnicode_Check(item)) {
                 Py_DECREF(seq);
                 PyErr_SetString(PyExc_TypeError, "expected a string");
                 goto failed;
             }
-            bytes = PyUnicode_AsUTF8String(item);
-            if (bytes == NULL) {
+
+            Py_ssize_t size;
+            const char *feature = PyUnicode_AsUTF8AndSize(item, &size);
+            if (feature == NULL) {
                 Py_DECREF(seq);
                 goto failed;
             }
-            feature = PyBytes_AS_STRING(bytes);
-            size = PyBytes_GET_SIZE(bytes);
             if (!raqm_add_font_feature(rq, feature, size)) {
                 Py_DECREF(seq);
-                Py_DECREF(bytes);
                 PyErr_SetString(PyExc_ValueError, "raqm_add_font_feature() failed");
                 goto failed;
             }
-            Py_DECREF(bytes);
         }
         Py_DECREF(seq);
     }
@@ -404,29 +409,31 @@ text_layout_fallback(
     const char *lang,
     GlyphInfo **glyph_info,
     int mask,
-    int color) {
-    int error, load_flags;
+    int color
+) {
+    int error, load_flags, i;
+    char *buffer = NULL;
     FT_ULong ch;
     Py_ssize_t count;
     FT_GlyphSlot glyph;
     FT_Bool kerning = FT_HAS_KERNING(self->face);
     FT_UInt last_index = 0;
-    int i;
 
     if (features != Py_None || dir != NULL || lang != NULL) {
         PyErr_SetString(
             PyExc_KeyError,
             "setting text direction, language or font features is not supported "
-            "without libraqm");
-    }
-    if (!PyUnicode_Check(string)) {
-        PyErr_SetString(PyExc_TypeError, "expected string");
-        return 0;
+            "without libraqm"
+        );
     }
 
-    count = 0;
-    while (font_getchar(string, count, &ch)) {
-        count++;
+    if (PyUnicode_Check(string)) {
+        count = PyUnicode_GET_LENGTH(string);
+    } else if (PyBytes_Check(string)) {
+        PyBytes_AsStringAndSize(string, &buffer, &count);
+    } else {
+        PyErr_SetString(PyExc_TypeError, "expected string or bytes");
+        return 0;
     }
     if (count == 0) {
         return 0;
@@ -445,7 +452,12 @@ text_layout_fallback(
     if (color) {
         load_flags |= FT_LOAD_COLOR;
     }
-    for (i = 0; font_getchar(string, i, &ch); i++) {
+    for (i = 0; i < count; i++) {
+        if (buffer) {
+            ch = buffer[i];
+        } else {
+            ch = PyUnicode_READ_CHAR(string, i);
+        }
         (*glyph_info)[i].index = FT_Get_Char_Index(self->face, ch);
         error = FT_Load_Glyph(self->face, (*glyph_info)[i].index, load_flags);
         if (error) {
@@ -462,7 +474,8 @@ text_layout_fallback(
                     last_index,
                     (*glyph_info)[i].index,
                     ft_kerning_default,
-                    &delta) == 0) {
+                    &delta
+                ) == 0) {
                 (*glyph_info)[i - 1].x_advance += PIXEL(delta.x);
                 (*glyph_info)[i - 1].y_advance += PIXEL(delta.y);
             }
@@ -486,17 +499,18 @@ text_layout(
     const char *lang,
     GlyphInfo **glyph_info,
     int mask,
-    int color) {
+    int color
+) {
     size_t count;
 #ifdef HAVE_RAQM
     if (have_raqm && self->layout_engine == LAYOUT_RAQM) {
-        count = text_layout_raqm(
-            string, self, dir, features, lang, glyph_info);
+        count = text_layout_raqm(string, self, dir, features, lang, glyph_info);
     } else
 #endif
     {
         count = text_layout_fallback(
-            string, self, dir, features, lang, glyph_info, mask, color);
+            string, self, dir, features, lang, glyph_info, mask, color
+        );
     }
     return count;
 }
@@ -518,7 +532,8 @@ font_getlength(FontObject *self, PyObject *args) {
     /* calculate size and bearing for a given string */
 
     if (!PyArg_ParseTuple(
-            args, "O|zzOz:getlength", &string, &mode, &dir, &features, &lang)) {
+            args, "O|zzOz:getlength", &string, &mode, &dir, &features, &lang
+        )) {
         return NULL;
     }
 
@@ -550,7 +565,18 @@ font_getlength(FontObject *self, PyObject *args) {
 }
 
 static int
-bounding_box_and_anchors(FT_Face face, const char *anchor, int horizontal_dir, GlyphInfo *glyph_info, size_t count, int load_flags, int *width, int *height, int *x_offset, int *y_offset) {
+bounding_box_and_anchors(
+    FT_Face face,
+    const char *anchor,
+    int horizontal_dir,
+    GlyphInfo *glyph_info,
+    size_t count,
+    int load_flags,
+    int *width,
+    int *height,
+    int *x_offset,
+    int *y_offset
+) {
     int position; /* pen position along primary axis, in 26.6 precision */
     int advanced; /* pen position along primary axis, in pixels */
     int px, py;   /* position of current glyph, in pixels */
@@ -558,8 +584,8 @@ bounding_box_and_anchors(FT_Face face, const char *anchor, int horizontal_dir, G
     int x_anchor, y_anchor;         /* offset of point drawn at (0, 0), in pixels */
     int error;
     FT_Glyph glyph;
-    FT_BBox bbox;                   /* glyph bounding box */
-    size_t i;                       /* glyph_info index */
+    FT_BBox bbox; /* glyph bounding box */
+    size_t i;     /* glyph_info index */
     /*
      * text bounds are given by:
      *   - bounding boxes of individual glyphs
@@ -654,9 +680,9 @@ bounding_box_and_anchors(FT_Face face, const char *anchor, int horizontal_dir, G
                     break;
                 case 'm':  // middle (ascender + descender) / 2
                     y_anchor = PIXEL(
-                        (face->size->metrics.ascender +
-                         face->size->metrics.descender) /
-                        2);
+                        (face->size->metrics.ascender + face->size->metrics.descender) /
+                        2
+                    );
                     break;
                 case 's':  // horizontal baseline
                     y_anchor = 0;
@@ -719,7 +745,7 @@ bad_anchor:
 static PyObject *
 font_getsize(FontObject *self, PyObject *args) {
     int width, height, x_offset, y_offset;
-    int load_flags;               /* FreeType load_flags parameter */
+    int load_flags; /* FreeType load_flags parameter */
     int error;
     GlyphInfo *glyph_info = NULL; /* computed text layout */
     size_t count;                 /* glyph_info length */
@@ -736,7 +762,8 @@ font_getsize(FontObject *self, PyObject *args) {
     /* calculate size and bearing for a given string */
 
     if (!PyArg_ParseTuple(
-            args, "O|zzOzz:getsize", &string, &mode, &dir, &features, &lang, &anchor)) {
+            args, "O|zzOzz:getsize", &string, &mode, &dir, &features, &lang, &anchor
+        )) {
         return NULL;
     }
 
@@ -758,7 +785,18 @@ font_getsize(FontObject *self, PyObject *args) {
         load_flags |= FT_LOAD_COLOR;
     }
 
-    error = bounding_box_and_anchors(self->face, anchor, horizontal_dir, glyph_info, count, load_flags, &width, &height, &x_offset, &y_offset);
+    error = bounding_box_and_anchors(
+        self->face,
+        anchor,
+        horizontal_dir,
+        glyph_info,
+        count,
+        load_flags,
+        &width,
+        &height,
+        &x_offset,
+        &y_offset
+    );
     if (glyph_info) {
         PyMem_Free(glyph_info);
         glyph_info = NULL;
@@ -767,12 +805,7 @@ font_getsize(FontObject *self, PyObject *args) {
         return NULL;
     }
 
-    return Py_BuildValue(
-        "(ii)(ii)",
-        width,
-        height,
-        x_offset,
-        y_offset);
+    return Py_BuildValue("(ii)(ii)", width, height, x_offset, y_offset);
 }
 
 static PyObject *
@@ -798,10 +831,10 @@ font_render(FontObject *self, PyObject *args) {
     unsigned char convert_scale;    /* scale factor for non-8bpp bitmaps */
     PyObject *image;
     Imaging im;
-    Py_ssize_t id;
     int mask = 0;  /* is FT_LOAD_TARGET_MONO enabled? */
     int color = 0; /* is FT_LOAD_COLOR enabled? */
-    int stroke_width = 0;
+    float stroke_width = 0;
+    int stroke_filled = 0;
     PY_LONG_LONG foreground_ink_long = 0;
     unsigned int foreground_ink;
     const char *mode = NULL;
@@ -815,14 +848,13 @@ font_render(FontObject *self, PyObject *args) {
     float y_start = 0;
     int width, height, x_offset, y_offset;
     int horizontal_dir; /* is primary axis horizontal? */
-    PyObject *max_image_pixels = Py_None;
 
     /* render string into given buffer (the buffer *must* have
        the right size, or this will crash) */
 
     if (!PyArg_ParseTuple(
             args,
-            "OO|zzOzizLffO:render",
+            "OO|zzOzfpzL(ff):render",
             &string,
             &fill,
             &mode,
@@ -830,11 +862,12 @@ font_render(FontObject *self, PyObject *args) {
             &features,
             &lang,
             &stroke_width,
+            &stroke_filled,
             &anchor,
             &foreground_ink_long,
             &x_start,
-            &y_start,
-            &max_image_pixels)) {
+            &y_start
+        )) {
         return NULL;
     }
 
@@ -871,34 +904,39 @@ font_render(FontObject *self, PyObject *args) {
 
     horizontal_dir = dir && strcmp(dir, "ttb") == 0 ? 0 : 1;
 
-    error = bounding_box_and_anchors(self->face, anchor, horizontal_dir, glyph_info, count, load_flags, &width, &height, &x_offset, &y_offset);
+    error = bounding_box_and_anchors(
+        self->face,
+        anchor,
+        horizontal_dir,
+        glyph_info,
+        count,
+        load_flags,
+        &width,
+        &height,
+        &x_offset,
+        &y_offset
+    );
     if (error) {
         PyMem_Del(glyph_info);
         return NULL;
     }
 
-    width += stroke_width * 2 + ceil(x_start);
-    height += stroke_width * 2 + ceil(y_start);
-    if (max_image_pixels != Py_None) {
-        if ((long long)(width > 1 ? width : 1) * (height > 1 ? height : 1) > PyLong_AsLongLong(max_image_pixels) * 2) {
-            PyMem_Del(glyph_info);
-            return Py_BuildValue("(ii)(ii)", width, height, 0, 0);
-        }
-    }
-
-    image = PyObject_CallFunction(fill, "s(ii)", strcmp(mode, "RGBA") == 0 ? "RGBA" : "L", width, height);
+    width += ceil(stroke_width * 2 + x_start);
+    height += ceil(stroke_width * 2 + y_start);
+    image = PyObject_CallFunction(fill, "ii", width, height);
     if (image == NULL) {
         PyMem_Del(glyph_info);
         return NULL;
     }
-    id = PyLong_AsSsize_t(PyObject_GetAttrString(image, "id"));
-    im = (Imaging)id;
+    PyObject *imagePtr = PyObject_GetAttrString(image, "ptr");
+    im = (Imaging)PyCapsule_GetPointer(imagePtr, IMAGING_MAGIC);
+    Py_XDECREF(imagePtr);
 
-    x_offset -= stroke_width;
-    y_offset -= stroke_width;
+    x_offset = round(x_offset - stroke_width);
+    y_offset = round(y_offset - stroke_width);
     if (count == 0 || width == 0 || height == 0) {
         PyMem_Del(glyph_info);
-        return Py_BuildValue("(ii)(ii)", width, height, x_offset, y_offset);
+        return Py_BuildValue("N(ii)", image, x_offset, y_offset);
     }
 
     if (stroke_width) {
@@ -910,10 +948,11 @@ font_render(FontObject *self, PyObject *args) {
 
         FT_Stroker_Set(
             stroker,
-            (FT_Fixed)stroke_width * 64,
+            (FT_Fixed)round(stroke_width * 64),
             FT_STROKER_LINECAP_ROUND,
             FT_STROKER_LINEJOIN_ROUND,
-            0);
+            0
+        );
     }
 
     /*
@@ -947,8 +986,8 @@ font_render(FontObject *self, PyObject *args) {
     }
 
     /* set pen position to text origin */
-    x = (-x_min + stroke_width + x_start) * 64;
-    y = (-y_max + (-stroke_width) - y_start) * 64;
+    x = round((-x_min + stroke_width + x_start) * 64);
+    y = round((-y_max + (-stroke_width) - y_start) * 64);
 
     if (stroker == NULL) {
         load_flags |= FT_LOAD_RENDER;
@@ -968,7 +1007,8 @@ font_render(FontObject *self, PyObject *args) {
         if (stroker != NULL) {
             error = FT_Get_Glyph(glyph_slot, &glyph);
             if (!error) {
-                error = FT_Glyph_Stroke(&glyph, stroker, 1);
+                error = stroke_filled ? FT_Glyph_StrokeBorder(&glyph, stroker, 0, 1)
+                                      : FT_Glyph_Stroke(&glyph, stroker, 1);
             }
             if (!error) {
                 FT_Vector origin = {0, 0};
@@ -1053,8 +1093,8 @@ font_render(FontObject *self, PyObject *args) {
             if (yy >= 0 && yy < im->ysize) {
                 /* blend this glyph into the buffer */
                 int k;
-                unsigned char v;
                 unsigned char *target;
+                unsigned int tmp;
                 if (color) {
                     /* target[RGB] returns the color, target[A] returns the mask */
                     /* target bands get split again in ImageDraw.text */
@@ -1065,34 +1105,83 @@ font_render(FontObject *self, PyObject *args) {
                 if (color && bitmap.pixel_mode == FT_PIXEL_MODE_BGRA) {
                     /* paste color glyph */
                     for (k = x0; k < x1; k++) {
-                        if (target[k * 4 + 3] < source[k * 4 + 3]) {
-                            /* unpremultiply BGRa to RGBA */
-                            target[k * 4 + 0] = CLIP8(
-                                (255 * (int)source[k * 4 + 2]) / source[k * 4 + 3]);
-                            target[k * 4 + 1] = CLIP8(
-                                (255 * (int)source[k * 4 + 1]) / source[k * 4 + 3]);
-                            target[k * 4 + 2] = CLIP8(
-                                (255 * (int)source[k * 4 + 0]) / source[k * 4 + 3]);
-                            target[k * 4 + 3] = source[k * 4 + 3];
+                        unsigned int src_alpha = source[k * 4 + 3];
+
+                        /* paste only if source has data */
+                        if (src_alpha > 0) {
+                            /* unpremultiply BGRa */
+                            int src_red =
+                                CLIP8((255 * (int)source[k * 4 + 2]) / src_alpha);
+                            int src_green =
+                                CLIP8((255 * (int)source[k * 4 + 1]) / src_alpha);
+                            int src_blue =
+                                CLIP8((255 * (int)source[k * 4 + 0]) / src_alpha);
+
+                            /* blend required if target has data */
+                            if (target[k * 4 + 3] > 0) {
+                                /* blend RGBA colors */
+                                target[k * 4 + 0] =
+                                    BLEND(src_alpha, target[k * 4 + 0], src_red, tmp);
+                                target[k * 4 + 1] =
+                                    BLEND(src_alpha, target[k * 4 + 1], src_green, tmp);
+                                target[k * 4 + 2] =
+                                    BLEND(src_alpha, target[k * 4 + 2], src_blue, tmp);
+                                target[k * 4 + 3] = CLIP8(
+                                    src_alpha +
+                                    MULDIV255(target[k * 4 + 3], (255 - src_alpha), tmp)
+                                );
+                            } else {
+                                /* paste unpremultiplied RGBA values */
+                                target[k * 4 + 0] = src_red;
+                                target[k * 4 + 1] = src_green;
+                                target[k * 4 + 2] = src_blue;
+                                target[k * 4 + 3] = src_alpha;
+                            }
                         }
                     }
                 } else if (bitmap.pixel_mode == FT_PIXEL_MODE_GRAY) {
                     if (color) {
                         unsigned char *ink = (unsigned char *)&foreground_ink;
                         for (k = x0; k < x1; k++) {
-                            v = source[k] * convert_scale;
-                            if (target[k * 4 + 3] < v) {
-                                target[k * 4 + 0] = ink[0];
-                                target[k * 4 + 1] = ink[1];
-                                target[k * 4 + 2] = ink[2];
-                                target[k * 4 + 3] = v;
+                            unsigned int src_alpha = source[k] * convert_scale;
+                            if (src_alpha > 0) {
+                                if (target[k * 4 + 3] > 0) {
+                                    target[k * 4 + 0] = BLEND(
+                                        src_alpha, target[k * 4 + 0], ink[0], tmp
+                                    );
+                                    target[k * 4 + 1] = BLEND(
+                                        src_alpha, target[k * 4 + 1], ink[1], tmp
+                                    );
+                                    target[k * 4 + 2] = BLEND(
+                                        src_alpha, target[k * 4 + 2], ink[2], tmp
+                                    );
+                                    target[k * 4 + 3] = CLIP8(
+                                        src_alpha +
+                                        MULDIV255(
+                                            target[k * 4 + 3], (255 - src_alpha), tmp
+                                        )
+                                    );
+                                } else {
+                                    target[k * 4 + 0] = ink[0];
+                                    target[k * 4 + 1] = ink[1];
+                                    target[k * 4 + 2] = ink[2];
+                                    target[k * 4 + 3] = src_alpha;
+                                }
                             }
                         }
                     } else {
                         for (k = x0; k < x1; k++) {
-                            v = source[k] * convert_scale;
-                            if (target[k] < v) {
-                                target[k] = v;
+                            unsigned int src_alpha = source[k] * convert_scale;
+                            if (src_alpha > 0) {
+                                target[k] =
+                                    target[k] > 0
+                                        ? CLIP8(
+                                              src_alpha +
+                                              MULDIV255(
+                                                  target[k], (255 - src_alpha), tmp
+                                              )
+                                          )
+                                        : src_alpha;
                             }
                         }
                     }
@@ -1113,18 +1202,12 @@ font_render(FontObject *self, PyObject *args) {
     if (bitmap_converted_ready) {
         FT_Bitmap_Done(library, &bitmap_converted);
     }
-    Py_DECREF(image);
     FT_Stroker_Done(stroker);
     PyMem_Del(glyph_info);
-    return Py_BuildValue("(ii)(ii)", width, height, x_offset, y_offset);
+    return Py_BuildValue("N(ii)", image, x_offset, y_offset);
 
 glyph_error:
-    if (im->destroy) {
-        im->destroy(im);
-    }
-    if (im->image) {
-        free(im->image);
-    }
+    Py_DECREF(image);
     if (stroker != NULL) {
         FT_Done_Glyph(glyph);
     }
@@ -1158,30 +1241,49 @@ font_getvarnames(FontObject *self) {
         return NULL;
     }
 
+    int *list_names_filled = PyMem_Malloc(num_namedstyles * sizeof(int));
+    if (list_names_filled == NULL) {
+        Py_DECREF(list_names);
+        FT_Done_MM_Var(library, master);
+        return PyErr_NoMemory();
+    }
+
+    for (unsigned int i = 0; i < num_namedstyles; i++) {
+        list_names_filled[i] = 0;
+    }
+
     name_count = FT_Get_Sfnt_Name_Count(self->face);
     for (i = 0; i < name_count; i++) {
         error = FT_Get_Sfnt_Name(self->face, i, &name);
         if (error) {
+            PyMem_Free(list_names_filled);
             Py_DECREF(list_names);
             FT_Done_MM_Var(library, master);
             return geterror(error);
         }
 
         for (j = 0; j < num_namedstyles; j++) {
-            if (PyList_GetItem(list_names, j) != NULL) {
+            if (list_names_filled[j]) {
                 continue;
             }
 
             if (master->namedstyle[j].strid == name.name_id) {
                 list_name = Py_BuildValue("y#", name.string, name.string_len);
+                if (list_name == NULL) {
+                    PyMem_Free(list_names_filled);
+                    Py_DECREF(list_names);
+                    FT_Done_MM_Var(library, master);
+                    return NULL;
+                }
                 PyList_SetItem(list_names, j, list_name);
+                list_names_filled[j] = 1;
                 break;
             }
         }
     }
 
+    PyMem_Free(list_names_filled);
     FT_Done_MM_Var(library, master);
-
     return list_names;
 }
 
@@ -1238,8 +1340,14 @@ font_getvaraxes(FontObject *self) {
 
             if (name.name_id == axis.strid) {
                 axis_name = Py_BuildValue("y#", name.string, name.string_len);
-                PyDict_SetItemString(list_axis, "name", axis_name ? axis_name : Py_None);
-                Py_XDECREF(axis_name);
+                if (axis_name == NULL) {
+                    Py_DECREF(list_axis);
+                    Py_DECREF(list_axes);
+                    FT_Done_MM_Var(library, master);
+                    return NULL;
+                }
+                PyDict_SetItemString(list_axis, "name", axis_name);
+                Py_DECREF(axis_name);
                 break;
             }
         }
@@ -1266,8 +1374,7 @@ font_setvarname(FontObject *self, PyObject *args) {
         return geterror(error);
     }
 
-    Py_INCREF(Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
 }
 
 static PyObject *
@@ -1288,12 +1395,17 @@ font_setvaraxes(FontObject *self, PyObject *args) {
     }
 
     num_coords = PyObject_Length(axes);
-    coords = (FT_Fixed*)malloc(num_coords * sizeof(FT_Fixed));
+    coords = (FT_Fixed *)malloc(num_coords * sizeof(FT_Fixed));
     if (coords == NULL) {
         return PyErr_NoMemory();
     }
     for (i = 0; i < num_coords; i++) {
-        item = PyList_GET_ITEM(axes, i);
+        item = PyList_GetItemRef(axes, i);
+        if (item == NULL) {
+            free(coords);
+            return NULL;
+        }
+
         if (PyFloat_Check(item)) {
             coord = PyFloat_AS_DOUBLE(item);
         } else if (PyLong_Check(item)) {
@@ -1301,10 +1413,12 @@ font_setvaraxes(FontObject *self, PyObject *args) {
         } else if (PyNumber_Check(item)) {
             coord = PyFloat_AsDouble(item);
         } else {
+            Py_DECREF(item);
             free(coords);
             PyErr_SetString(PyExc_TypeError, "list must contain numbers");
             return NULL;
         }
+        Py_DECREF(item);
         coords[i] = coord * 65536;
     }
 
@@ -1314,15 +1428,16 @@ font_setvaraxes(FontObject *self, PyObject *args) {
         return geterror(error);
     }
 
-    Py_INCREF(Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
 }
 #endif
 
 static void
 font_dealloc(FontObject *self) {
     if (self->face) {
+        MUTEX_LOCK(&ft_library_mutex);
         FT_Done_Face(self->face);
+        MUTEX_UNLOCK(&ft_library_mutex);
     }
     if (self->font_bytes) {
         PyMem_Free(self->font_bytes);
@@ -1341,7 +1456,8 @@ static PyMethodDef font_methods[] = {
     {"setvarname", (PyCFunction)font_setvarname, METH_VARARGS},
     {"setvaraxes", (PyCFunction)font_setvaraxes, METH_VARARGS},
 #endif
-    {NULL, NULL}};
+    {NULL, NULL}
+};
 
 static PyObject *
 font_getattr_family(FontObject *self, void *closure) {
@@ -1398,43 +1514,20 @@ static struct PyGetSetDef font_getsetters[] = {
     {"x_ppem", (getter)font_getattr_x_ppem},
     {"y_ppem", (getter)font_getattr_y_ppem},
     {"glyphs", (getter)font_getattr_glyphs},
-    {NULL}};
+    {NULL}
+};
 
 static PyTypeObject Font_Type = {
-    PyVarObject_HEAD_INIT(NULL, 0) "Font",
-    sizeof(FontObject),
-    0,
-    /* methods */
-    (destructor)font_dealloc, /* tp_dealloc */
-    0,                        /* tp_print */
-    0,                        /*tp_getattr*/
-    0,                        /*tp_setattr*/
-    0,                        /*tp_compare*/
-    0,                        /*tp_repr*/
-    0,                        /*tp_as_number */
-    0,                        /*tp_as_sequence */
-    0,                        /*tp_as_mapping */
-    0,                        /*tp_hash*/
-    0,                        /*tp_call*/
-    0,                        /*tp_str*/
-    0,                        /*tp_getattro*/
-    0,                        /*tp_setattro*/
-    0,                        /*tp_as_buffer*/
-    Py_TPFLAGS_DEFAULT,       /*tp_flags*/
-    0,                        /*tp_doc*/
-    0,                        /*tp_traverse*/
-    0,                        /*tp_clear*/
-    0,                        /*tp_richcompare*/
-    0,                        /*tp_weaklistoffset*/
-    0,                        /*tp_iter*/
-    0,                        /*tp_iternext*/
-    font_methods,             /*tp_methods*/
-    0,                        /*tp_members*/
-    font_getsetters,          /*tp_getset*/
+    PyVarObject_HEAD_INIT(NULL, 0).tp_name = "Font",
+    .tp_basicsize = sizeof(FontObject),
+    .tp_dealloc = (destructor)font_dealloc,
+    .tp_methods = font_methods,
+    .tp_getset = font_getsetters,
 };
 
 static PyMethodDef _functions[] = {
-    {"getfont", (PyCFunction)getfont, METH_VARARGS | METH_KEYWORDS}, {NULL, NULL}};
+    {"getfont", (PyCFunction)getfont, METH_VARARGS | METH_KEYWORDS}, {NULL, NULL}
+};
 
 static int
 setup_module(PyObject *m) {
@@ -1512,10 +1605,9 @@ PyInit__imagingft(void) {
 
     static PyModuleDef module_def = {
         PyModuleDef_HEAD_INIT,
-        "_imagingft", /* m_name */
-        NULL,         /* m_doc */
-        -1,           /* m_size */
-        _functions,   /* m_methods */
+        .m_name = "_imagingft",
+        .m_size = -1,
+        .m_methods = _functions,
     };
 
     m = PyModule_Create(&module_def);
@@ -1524,5 +1616,9 @@ PyInit__imagingft(void) {
         return NULL;
     }
 
+#ifdef Py_GIL_DISABLED
+    PyUnstable_Module_SetGIL(m, Py_MOD_GIL_NOT_USED);
+#endif
+
     return m;
 }
